--------------------------------------------------------------------------------
--
--  Murphi Model of the Location Privacy Protocol 1
--
--------------------------------------------------------------------------------
--
--  version:      1.0
--
--  written by:   Mugdha Lakhani
--  date:         Feb 2011
--  affiliation:  Stanford University
--
--------------------------------------------------------------------------------
--
--  the following three steps of the protocol are modeled:
--
--   1. A->B: f(a)
--   2. B->A: g(f(a),b)         -- A assumes it is talking to B
--
--   A: initiator, B: reponder
--
--------------------------------------------------------------------------------


--------------------------------------------------------------------------------
-- constants, types and variables
--------------------------------------------------------------------------------


-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
const

  NumAttackers:   2;   -- number of attackers
  NumVictims:	  1;   -- number of victims, fixed, not a parameter.
  NetworkSize:     10;   -- max. number of outstanding messages in network
MaxSteps:		   60;  -- max. number of steps before which error condition should be found
Threshold:        0.0001; -- max. probability theshold of error condition to occur
MaxLocations:     64800;
NumConversations: NumAttackers;
  
-- flags, for possible future use
Encryption_OK:    1;
Authentication_OK: 1;
Hashing_OK:       1;

-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
type

   AttackerId:  scalarset (NumAttackers);   -- identifiers
  VictimId:	scalarset (NumVictims);	-- identifiers
   AgentId:      union {AttackerId, VictimId};
Location:     1..MaxLocations; 
  prob:         real(10, 99); 

  CompromisedLoc: record
	loc_found:boolean;			 -- location compromised or not
	--found_with_prob: prob;		 -- what was the probability of finding the location,
								-- not required, can be approximated assuming every rule has equal chance of
								-- getting fired, and every agent is equally likely to run 1 instance of the protocol.
   --who_leaked: AgentId			useful to know when dealing with multiple roles per agent
  end;



  MessageType : enum {           -- different types of messages
    From_Init,				     --  f(a):   initiator's address
    From_Resp                   --  {g(f(a),b)}: responder's address
  };

  Message : record
    source:   AgentId;           -- source of message
    dest:     AgentId;           -- intended destination of message
    key:      AgentId;           -- key used for encryption
    mType:    MessageType;       -- type of message
	f_a:      Location;           -- sender's location
    g_a_b:    Location;          -- initiator's loc - responder's loc
  end;

  -- remark: addresses are encoded in the message by their
  --         agent's identifier only. They can be distinguished by their po-
  --         sition and the type of the message

  AgentStates : enum {
    I_SLEEP,                     -- state after initialization, start state
    I_WAIT,                      -- waiting for response from responder
    I_RECV,                       -- initiator receives g(f(a),b)
	I_KNOW,						 -- I knows b
	-- responder states
	R_SLEEP,				     -- state after initialization, start state
	R_RECV,						 -- received f(a)
	R_SEND,						 -- sent g(f(a), b)
	R_KNOW						 -- R knows a
  };             

	Roles: enum {
		initiator,
		responder
  };
	Attacker:  record
		role: Roles; 				-- which role the agent is in
		state:AgentStates;      -- which state the agent is in
		other: AgentId;			-- other person in the conversation
		location: Location;
		locs_to_explore: 1..MaxLocations/NumAttackers;

	  end;

Conversation: record
	role: Roles;						-- victim role in this conversation conversation
	other: AgentId;					-- victim could be talking to other victims, or to attackers
	state: AgentStates;				-- victim has a different state for every conversation
end;

Victim: record
	location: Location;
	conversation_list: array[AttackerId] of Conversation;
end;

-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
var                                         
  net: multiset[NetworkSize] of Message;   --  network
  atkr: array[AttackerId] of Attacker;     --  initiators
  vic: array[VictimId] of Victim;          --  responders
  compromised_locs: array[VictimId] of CompromisedLoc;

--------------------------------------------------------------------------------
-- functions and procedures
--------------------------------------------------------------------------------
-- calc_prob takes two arguments: attacker_id and state s, and returns the
-- probability that attacker attacker_id will go to state s.

function calc_prob(i : AttackerId; c : AgentStates) : prob;
var rho : prob;
begin
alias remaining : atkr[i].locs_to_explore do
put "\n attacker ";
put i;
put "  to state ";
put c;
put "  with ";
put remaining;
put "   locations to explore\n";

switch atkr[i].state

case I_SLEEP:
	put "case I_SLEEP";
	if c = I_WAIT then return 1.0;
	else 
		put "returning 0\n";
		return 0.0;
	endif;

case I_WAIT:
	put "case I_WAIT";
	if c = I_RECV then
		put "not returning 0";
		return (1.0 - (1.0/remaining));
	elsif c = I_KNOW then
			put "not returning 0";
			return 1.0/remaining; 
	else 
		put "returning 0";
		return 0.0;
	endif;

case I_RECV:
	put "case I_RECV\n";
	if c = I_SLEEP then 
		put "I_RECV to I_SLEEP";
		return 1.0; 
	else
		put "I_RECV to OTHER";
		put "returning 0";
		return 0.0;
	endif;
case I_KNOW:
	put "case I_KNOW";
	if c = I_KNOW then return 1.0;
	else
		put "returning 0";
		return 0.0;
	endif;


case R_SLEEP:
	put "case R_SLEEP";
	if c = R_RECV then return (1.0);
	else 
		put "returning 0";
		return 0.0;
	endif
case R_RECV:
	put "case R_RECV";
	if c = R_SEND then return (1.0);
	else 
		put "returning 0";
		return 0.0;
	endif
case R_SEND:
	put "case R_SEND";
	if c = R_SLEEP then return (1.0);
	else 
		put "returning 0";
		return 0.0;
	endif
case R_KNOW:
	put "case R_KNOW";
	if c = R_KNOW then return 1.0;
	else
		put "returning 0";
		return 0.0;
	endif;

endswitch;
endalias;
end;

function send_msg(mType : MessageType; source : AgentId; dest: AgentId; f_a: Location; g_a_b: Location): boolean;
var outM : Message;	-- outgoing message
begin

outM.source := source;
outM.dest := dest;
outM.mType := mType;
if mType = From_Init then
outM.key := source;
outM.f_a := f_a;
else outM.key := dest;
outM.g_a_b := g_a_b;
endif;
if (multisetcount(l:net, true) < NetworkSize) then
multisetadd(outM, net);
return true;
else return false;
endif
end;

function recv_msg(mType: MessageType;source: AgentId; dest: AgentId): boolean;
begin
if multisetcount(j:net,
				 net[j].source = source & 
				 net[j].dest = dest & 
				 net[j].mType = mType) = 0 then 
return false;
else 

	multisetremovepred(o:net, net[o].source = source &
				   net[o].dest = dest &
				   net[o].mType = mType);
	return true;
endif
end;


procedure step(i: AttackerId; c: AgentStates);
var change_state : boolean;
begin
alias attacker : atkr[i]; victim : vic[atkr[i].other]; viclist : vic[atkr[i].other].conversation_list do
put "step";
switch attacker.state
case I_SLEEP:
	if c = I_WAIT then 
		--attacker sends message
		change_state := send_msg(From_Init,i, attacker.other, attacker.location, 1);
		if change_state then 
			attacker.state := c;
			put "attacker state update to ";
			put c;
			put "\n";
		endif;
		-- victim receives message
		change_state := recv_msg(From_Init,i, attacker.other);
		if change_state then
			--find this conversation and update state
		for p: AttackerId do
				if (viclist[p].other = i) then viclist[p].state := R_RECV;
				endif
			endfor
		endif
	elsif c = I_KNOW then
		attacker.state := I_KNOW;
		put "attacker state update to ";
		put I_KNOW;
		put "\n";
		compromised_locs[attacker.other].loc_found := true;
	endif
	
case I_WAIT:
	if c = I_RECV then
		-- victim sends message
		change_state := send_msg(From_Resp,attacker.other,i, 1, victim.location); 
		if change_state then 
			--find this conversation and update state
			for p: AttackerId do
				if (viclist[p].other = i) then viclist[p].state := R_SEND;
				endif;
			endfor;
		endif;
		-- attacker receives message
		change_state := recv_msg(From_Resp, attacker.other, i);
		if change_state then 
			attacker.state := I_RECV;
			attacker.locs_to_explore := attacker.locs_to_explore - 1;
			put "2";
			put attacker.locs_to_explore;
		endif

		
	elsif c = I_KNOW then
		attacker.state := I_KNOW;
		compromised_locs[attacker.other].loc_found := true;
	endif
case I_RECV:
	if c = I_SLEEP then
		attacker.state := I_SLEEP;
		for p: AttackerId do
			if (viclist[p].other = i) then viclist[p].state := R_SLEEP;
			endif;
		endfor;
	elsif c = I_KNOW then 
		attacker.state := I_KNOW;
		compromised_locs[attacker.other].loc_found := true;
	endif
case R_SLEEP:
	if c = R_RECV then
		--victim sends message
		change_state := send_msg(From_Init, attacker.other, i,victim.location, 1);
		if change_state then
			for p: AttackerId do
				if (viclist[p].other = i) then viclist[p].state := I_WAIT;
				endif;
			endfor;
		endif;
		--attacker receives message
		change_state := recv_msg(From_Init, attacker.other, i);
		if change_state then attacker.state := R_RECV;
		endif

	elsif c = R_KNOW then 
		attacker.state := R_KNOW;
		compromised_locs[attacker.other].loc_found := true;
	endif;

case R_RECV:
	if c = R_SEND then
		--attacker sends Message
		change_state := send_msg(From_Resp, i, attacker.other, 1, attacker.location);
		if change_state then attacker.state := R_SEND;
		endif;
		--victim receives message
		change_state := recv_msg(From_Resp, i, attacker.other);
		if change_state then
			for p: AttackerId do
				if (viclist[p].other = i) then viclist[p].state := I_RECV;
				endif
			endfor;
		endif;

	elsif c = R_KNOW then 
		attacker.state := R_KNOW;
		compromised_locs[attacker.other].loc_found := true;
	endif;

case R_SEND:
	if c = R_SLEEP then
		attacker.state := R_SLEEP;
		for p: AttackerId do
			if (viclist[p].other = i) then viclist[p].state := I_SLEEP;
			endif
		endfor
	elsif c = R_KNOW then 
		attacker.state := R_KNOW;
		compromised_locs[attacker.other].loc_found := true;
	endif;
 endswitch;
endalias;
end;


--------------------------------------------------------------------------------
-- rules
--------------------------------------------------------------------------------
ruleset attacker: AttackerId do
	ruleset next_state : AgentStates do
		rule "next"
		    calc_prob(attacker, next_state) ==>
		begin
			step(attacker, next_state);
		end;
	end;
end;

--------------------------------------------------------------------------------
-- startstate
--------------------------------------------------------------------------------

startstate


  -- initialize attackers except other field
  undefine atkr;
  for i: AttackerId do
    atkr[i].role     := initiator;
	atkr[i].location := 1;			-- this location can be anything, attacker does not worry about revealing location
	atkr[i].state := I_SLEEP;
	atkr[i].locs_to_explore := MaxLocations/NumAttackers;

  end;

  -- initialize victims except other field
  undefine vic;

  for i: VictimId do
		vic[i].location := 1;  -- since attacker starts to check from a very large number, they do not try 1 soon
		for k: AttackerId do
			vic[i].conversation_list[k].role := responder;
			vic[i].conversation_list[k].other := i;
			vic[i].conversation_list[k].state := R_SLEEP;
		endfor; 
  endfor;


-- initialize the other fields of attackers and victims, note that there is just one victim
-- note that because AttackerId and VictimId are scalarsets, I figured out this way of assigning all attackers to one victim
-- with the exceptional case of first Attacker who is in the role of a responder.

for i: AttackerId do
	for j: VictimId do
		atkr[i].other := j;
		vic[j].conversation_list[i].other := i;
/*		-- exceptional condition for the first attacker and first victim conversation
		if (i = j) then
			atkr[i].role := responder;
			atkr[i].state := R_SLEEP;
			vic[j].conversation_list[i].role := initiator;
			vic[j].conversation_list[i].state := I_SLEEP;
		endif;
*/
	endfor;
endfor;


-- initialize compromised_locs
	undefine compromised_locs;
	for i: VictimId do
	compromised_locs[i].loc_found := false;
	endfor;
	 
  -- initialize network 
  undefine net;
endstartstate;



--------------------------------------------------------------------------------
-- invariants
--------------------------------------------------------------------------------
--invariant "less than networksize outstanding messages in network"
--multisetcount (l:net, true) < NetworkSize;

--------------------------------------------------------------------------------
-- property
--------------------------------------------------------------------------------

pctl
[
 true 
 UNTIL <= 7
 (exists v: VictimId  do (compromised_locs[v].loc_found) end)
 ] <= 0.001;





