--------------------------------------------------------------------------------
--
--  Murphi Model of the Location Privacy Protocol 1
--
--------------------------------------------------------------------------------
--
--  version:      1.0
--
--  written by:   Mugdha Lakhani
--  date:         Feb 2011
--  affiliation:  Stanford University
--
--------------------------------------------------------------------------------
--
--  the following three steps of the protocol are modeled:
--
--   1. B-> S: g(b)
--   2. A-> S: f(a)       
--   3. S-> A: h(g(b),f(a))
--   A: initiator, B: reponder; S: server
--------------------------------------------------------------------------------


--------------------------------------------------------------------------------
-- constants, types and variables
--------------------------------------------------------------------------------


-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
const

  NumAttackers:   2;   -- number of attackers
  NumVictims:	  1;   -- number of victims, fixed, not a parameter.
  NumServers:      1;   -- number of server, fixed, not a parameter.
  NetworkSize:     10;   -- max. number of outstanding messages in network
  MaxSteps:		   60;  -- max. number of steps before which error condition should be found
  Threshold:        0.0001; -- max. probability theshold of error condition to occur
  MaxLocations:     64800;
  NumConversations: NumAttackers;
  
-- flags, for possible future use
Encryption_OK:    1;
Authentication_OK: 1;
Hashing_OK:       1;

-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
type

  AttackerId:  scalarset (NumAttackers);   -- identifiers
  VictimId:	scalarset (NumVictims);	-- identifiers
  ServerId: scalarset(NumServers);
AgentId:      union {AttackerId, VictimId, ServerId};
   Location:     1..MaxLocations; 
   prob:         real(10, 99); 
   loc_found:boolean;					   -- location of victim compromised or not


  MessageType : enum {           -- different types of messages
    From_Init,				     -- f_a:   initiator's address
	From_Resp,					-- g_b: responder's address
	From_Server					 -- h_a_b : difference of the locations
  };

  Message : record
    source:   AgentId;           -- source of message
    dest:     AgentId;           -- intended destination of message
    key:      AgentId;           -- key used for encryption
    mType:    MessageType;       -- type of message
	f_a:      Location;          -- initiator's location
    g_b:    Location;            -- responder's loc
	h_a_b:	Location;			 -- diff of locations
  end;


  AgentStates : enum {
    I_SLEEP,                     -- state after initialization, start state
    I_WAIT,                      -- waiting for response from server
    I_RECV,                       -- initiator receives h_a_b
	I_KNOW,						 -- I knows b
	-- responder states
	R_SLEEP,				     -- state after initialization, start state
	R_SEND,						 -- sent g(b)
	R_KNOW,						 -- R knows a
	-- server states
	S_SLEEP,					 -- state after initialization, start state
	S_RECV1,					 -- after receiving g_b
	S_RECV2,					 -- after receiving f_a
 	S_SEND						 -- after sending h_a_b
  };             

	Roles: enum {
		initiator,
		responder,
		server
  };
	Attacker:  record
		role: Roles; 				-- which role the agent is in
		state:AgentStates;      -- which state the agent is in
		other: AgentId;			-- other person in the conversation
		server: AgentId;
		location: Location;
		locs_to_explore: 1..MaxLocations/NumAttackers;
	  end;

Conversation: record
	role: Roles;						-- victim role in this conversation conversation
	other: AgentId;					-- victim could be talking to other victims, or to attackers
	state: AgentStates;				-- victim has a different state for every conversation
end;

Victim: record
	location: Location;
	conversation_list: array[AttackerId] of Conversation;
end;

Server: record
	conversation_list: array [AttackerId] of Conversation;
end;

-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
var                                         
  net: multiset[NetworkSize] of Message;   --  network
  atkr: array[AttackerId] of Attacker;     --  initiators
  vic: array[VictimId] of Victim;          --  responders
  serv: array[ServerId] of Server;	       -- server
  compromised_locs : array[VictimId] of loc_found;
--------------------------------------------------------------------------------
-- functions and procedures
--------------------------------------------------------------------------------
-- calc_prob takes two arguments: attacker_id and state s, and returns the
-- probability that attacker attacker_id will go to state s.

function calc_prob(i : AttackerId; c : AgentStates) : prob;
begin
alias remaining : atkr[i].locs_to_explore do

switch atkr[i].state

case I_SLEEP:
	if c = I_WAIT then return 1.0;
	else 
		return 0.0;
	endif;

case I_WAIT:
	put "case I_WAIT";
	if c = I_RECV then
		put "not returning 0";
		return (1.0 - (1.0/remaining));
	elsif c = I_KNOW then
			put "not returning 0";
			return 1.0/remaining; 
	else 
		put "returning 0";
		return 0.0;
	endif;

case I_RECV:
	put "case I_RECV\n";
	if c = I_SLEEP then 
		put "I_RECV to I_SLEEP";
		return 1.0; 
	else
		put "I_RECV to OTHER";
		put "returning 0";
		return 0.0;
	endif;
case I_KNOW:
	put "case I_KNOW";
	if c = I_KNOW then return 1.0;
	else
		put "returning 0";
		return 0.0;
	endif;

endswitch;
endalias;
end;

function send_msg(mType : MessageType; source : AgentId; dest: AgentId; f_a: Location; g_b: Location; h_a_b: Location): boolean;
var outM : Message;	-- outgoing message
begin

outM.source := source;
outM.dest := dest;
outM.mType := mType;
if mType = From_Resp then
	outM.key := dest;		-- only server can read it
	outM.g_b := g_b;

elsif mType = From_Init then
	outM.key := dest;
	outM.f_a := f_a;

elsif mType = From_Server then
	outM.key := dest;
	outM.h_a_b := h_a_b;
endif;

if (multisetcount(l:net, true) < NetworkSize) then
	multisetadd(outM, net);
	return true;
else return false;
endif;
end;

function recv_msg(mType: MessageType;source: AgentId; dest: AgentId): boolean;
begin
if multisetcount(j:net,
				 net[j].source = source & 
				 net[j].dest = dest & 
				 net[j].mType = mType) = 0 then 
return false;
else 

	multisetremovepred(o:net, net[o].source = source &
				   net[o].dest = dest &
				   net[o].mType = mType);
	return true;
endif
end;


procedure step(i: AttackerId; c: AgentStates);
var change_state : boolean;
begin
alias attacker : atkr[i]; victim : vic[atkr[i].other];server: serv[atkr[i].server]; viclist :vic[atkr[i].other].conversation_list;
				 servlist: serv[atkr[i].server].conversation_list do
put "step";
switch attacker.state
case I_SLEEP:
	if c = I_WAIT then 
		--victim sends message
		change_state := send_msg(From_Resp,attacker.other, attacker.server, 1, victim.location, 1);
		if change_state then 
			--find this conversation and update state in servlist
			for p: AttackerId do
				if (viclist[p].other = i) then viclist[p].state := R_SEND;
				endif;
			endfor;
		endif;
		-- server receives message
		change_state := recv_msg(From_Resp,attacker.other, attacker.server);
		if change_state then
			--find this conversation and update state in servlist
		for p: AttackerId do
				if (servlist[p].other = i) then servlist[p].state := S_RECV1;
				endif;
			endfor;
		endif;
	
	   -- now attacker sends a message
		change_state := send_msg(From_Init,i, attacker.server, attacker.location, 1, 1);
		if change_state then 
			attacker.state := I_WAIT;
		endif;
	   -- server receives message
		change_state := recv_msg(From_Init,i, attacker.server);
		if change_state then
			--find this conversation and update state in servlist
			for p: AttackerId do
				if (servlist[p].other = i) then servlist[p].state := S_RECV2;
				endif;
			endfor;
		endif;
	endif;
		
case I_WAIT:
	if c = I_RECV then
		-- server sends message
		change_state := send_msg(From_Server,attacker.server,i, 1,1,victim.location); 
						-- does not matter what loc you send because we are not comparing
		if change_state then 
			--find this conversation and update state
			for p: AttackerId do
				if (servlist[p].other = i) then servlist[p].state := S_SEND;
				endif;
			endfor;
		endif;
		-- attacker receives message
		change_state := recv_msg(From_Server, attacker.server, i);
		if change_state then 
			attacker.state := I_RECV;
			attacker.locs_to_explore := attacker.locs_to_explore - 1;
			put "2";
			put attacker.locs_to_explore;
		endif
		
	elsif c = I_KNOW then
		attacker.state := I_KNOW;
		compromised_locs[attacker.other] := true;
	endif
case I_RECV:
	if c = I_SLEEP then
		attacker.state := I_SLEEP;
		for p: AttackerId do
			if (viclist[p].other = i) then viclist[p].state := R_SLEEP;
											servlist[p].state := S_SLEEP;
			endif;
		endfor;
	endif
 endswitch;
endalias;
end;


--------------------------------------------------------------------------------
-- rules
--------------------------------------------------------------------------------
ruleset attacker: AttackerId do
	ruleset next_state : AgentStates do
		rule "next"
		    calc_prob(attacker, next_state) ==>
		begin
			step(attacker, next_state);
		end;
	end;
end;

--------------------------------------------------------------------------------
-- startstate
--------------------------------------------------------------------------------

startstate


  -- initialize attackers except other field
  undefine atkr;
  for i: AttackerId do
    atkr[i].role     := initiator;
	atkr[i].location := 1;			-- this location can be anything, attacker does not worry about revealing location
	atkr[i].state := I_SLEEP;
	atkr[i].locs_to_explore := MaxLocations/NumAttackers;
  end;

  -- initialize victims except other field
  undefine vic;
	for j:VictimId do
		for k: AttackerId do
			vic[j].location := 1;  -- since attacker starts to check from a very large number, they do not try 1 soon
			vic[j].conversation_list[k].role := responder;
			vic[j].conversation_list[k].other := k;
			vic[j].conversation_list[k].state := R_SLEEP;
		endfor; 
	endfor;

-- initialize the other fields of attackers and victims

-- initialize other field of attacker
for i: AttackerId do
	for j: VictimId do
		atkr[i].other := j;
	endfor;
endfor;

-- initialize server field of attacker
for i: AttackerId do
for j: ServerId do
atkr[i].server := j;
endfor;
endfor;

-- initialize Server
for i: ServerId do
	for j:AttackerId do
		serv[i].conversation_list[j].state := S_SLEEP;
		serv[i].conversation_list[j].other := j;
		serv[i].conversation_list[j].role := server;
	endfor;
endfor;
	

-- initialize compromised_locs
	undefine compromised_locs;
for i: VictimId do
	compromised_locs[i] := false;
endfor;

	 
  -- initialize network 
  undefine net;
endstartstate;



--------------------------------------------------------------------------------
-- invariants
--------------------------------------------------------------------------------
--invariant "less than networksize outstanding messages in network"
--multisetcount (l:net, true) < NetworkSize;

--------------------------------------------------------------------------------
-- property
--------------------------------------------------------------------------------

pctl
[
 true 
 UNTIL <= 6
 (exists v: VictimId  do (compromised_locs[v]) end)
 ] <= 0.001;





