--------------------------------------------------------------------------------
--
--  Murphi Model of the Location Privacy Protocol 1
--
--------------------------------------------------------------------------------
--
--  version:      1.0
--
--  written by:   Mugdha Lakhani
--  date:         Feb 2011
--  affiliation:  Stanford University
--
--------------------------------------------------------------------------------
--
--  the following three steps of the protocol are modeled:
--
--   1. A->B: f(a)
--   2. B->A: g(f(a),b)         -- A assumes it is talking to B
--
--   A: initiator, B: reponder
--
--------------------------------------------------------------------------------


--------------------------------------------------------------------------------
-- constants, types and variables
--------------------------------------------------------------------------------


-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
const

  NumAttackers:   2;   -- number of attackers
  NumVictims:	  1;   -- number of victims, fixed, not a parameter.
  NetworkSize:     10;   -- max. number of outstanding messages in network
  MaxSteps:		   60;  -- max. number of steps before which error condition should be found
	Threshold:        0.0001; -- max. probability theshold of error condition to occur
	MaxLocations:     64800;
	NumConversations: NumAttackers;

-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
type

   AttackerId:  1..NumAttackers;   -- identifiers
AgentId:     0..NumAttackers;		-- 1 victim and n attackers.
Location:     1..MaxLocations; 
  prob:         real(10, 99); 



  MessageType : enum {           -- different types of messages
    From_Init,				     --  f(a):   initiator's address
    From_Resp                    --  g_b: responder's address
  };

  Message : record
    source:   AgentId;           -- source of message
    dest:     AgentId;           -- intended destination of message
    key:      AgentId;           -- key used for encryption
    mType:    MessageType;       -- type of message
	f_a:      Location;          -- sender's location
    g_a_b:    Location;            -- initiator's loc - responder's loc
  end;

  -- remark: addresses are encoded in the message by their
  --         agent's identifier only. They can be distinguished by their po-
  --         sition and the type of the message

  AgentStates : enum {
    I_SLEEP,                     -- state after initialization, start state
    I_WAIT,                      -- waiting for response from responder
    I_RECV,                       -- initiator receives g(f(a),b)
	I_KNOW,						 -- I knows b
	-- responder states
	R_SLEEP,				     -- state after initialization, start state
	R_RECV,						 -- state after receiving first msg from initiator
	R_SEND,						 -- sent g(f(a), b)
	R_KNOW						 -- R knows a

  };             

	Roles: enum {
		initiator,
		responder
  };
	Attacker:  record
		role: Roles; 				-- which role the agent is in
		state:AgentStates;      -- which state the agent is in
		other: AgentId;			-- other person in the conversation
		locs_to_explore: 1..MaxLocations/NumAttackers;
		location: Location;
		curprob: prob;			-- probability of being in current state
	  end;

Conversation: record
	role: Roles;						-- victim role in this conversation conversation
	other: AgentId;					-- victim could be talking to other victims, or to attackers
	state: AgentStates;				-- victim has a different state for every conversation
end;

Victim: record
	location: Location;
	conversation_list: array[AttackerId] of Conversation;
end;

-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
var                                         
	net: multiset[NetworkSize] of Message;   --  network
	atkr: array[AttackerId] of Attacker;     --  attacker
	vic: Victim;							 --  victim
	loc_found:boolean;			 -- location compromised or not
	found_with_prob: prob;		 -- what was the probability of finding the location

--------------------------------------------------------------------------------
-- functions and procedures
--------------------------------------------------------------------------------
-- calc_prob takes two arguments: attacker_id and state s, and returns the
-- probability that attacker attacker_id will go to state s.

function calc_prob(i : AttackerId; c : AgentStates) : prob;
var beta : prob;
begin
alias curprob : atkr[i].curprob do
beta := 1.0 / atkr[i].locs_to_explore;

switch atkr[i].state

case I_SLEEP:
	if c = I_WAIT then return (curprob);
	else return 0.0;
	endif

case I_WAIT:
	if c = I_RECV then
		return curprob * (1.0 - beta);
	elsif c = I_KNOW then
		return curprob * beta; 
	else return 0.0;
	endif

case I_RECV:
	if c = I_SLEEP then return curprob; 
	else return 0.0;
	endif
case I_KNOW:
	if c = I_KNOW then return curprob;
	else return 0.0;
	endif;
case R_SLEEP:
	if c = R_SEND then return curprob;
	else return 0.0;
	endif;
case R_SEND:
	if c = R_RECV then return curprob;
	else return 0.0;
	endif;
case R_RECV:
	if c = R_SLEEP then return curprob;
	else return 0.0;
	endif;
case R_KNOW:
	return 0.0;

endswitch;
endalias;
end;
-- function send_msg creates a message intended for an agent and adds it to the network
function send_msg(mType : MessageType; source : AgentId; dest: AgentId; f_a: Location; g_a_b: Location): boolean;
var outM : Message;	-- outgoing message
begin

outM.source := source;
outM.dest := dest;
outM.mType := mType;
if mType = From_Init then
outM.key := source;
outM.f_a := f_a;
else outM.key := dest;
outM.g_a_b := g_a_b;
endif;
if (multisetcount(l:net, true) < NetworkSize) then
multisetadd(outM, net);
return true;
else return false;
endif
end;

-- function recv_msg removes a certain msg from the network
function recv_msg(mType: MessageType;source: AgentId; dest: AgentId): boolean;
begin
if multisetcount(j:net,
				 net[j].source = source & 
				 net[j].dest = dest & 
				 net[j].mType = mType) = 0 then 
return false;
else 

	multisetremovepred(o:net, net[o].source = source &
				   net[o].dest = dest &
				   net[o].mType = mType);
	return true;
endif
end;

-- procedure step carries out the actual transition of an agent to a state when feasible
procedure step(i: AttackerId; c: AgentStates);
var change_state : boolean;
var new_prob : prob;
begin
alias attacker : atkr[i]; victim : vic; viclist : vic.conversation_list do
put "step\n";
new_prob := calc_prob(i, c);
put "prob of attacker  ";
put i;
put "  going to state";
put c;
put "  is ";
put new_prob;
put "\n";
switch attacker.state
case I_SLEEP:
	if c = I_WAIT then 
		--attacker sends message
		change_state := send_msg(From_Init,i, attacker.other, attacker.location, 1);
		put change_state;
		if change_state then 
			attacker.state := c;
			attacker.curprob := new_prob;
		endif;
		-- victim receives message
		change_state := recv_msg(From_Init,i, attacker.other);
		put change_state;
		if change_state then
			--find this conversation and update state
		for p: AttackerId do
				if (viclist[p].other = i) then viclist[p].state := R_SEND;
				endif
			endfor
		endif
	elsif c = I_KNOW then
		attacker.state := I_KNOW;
		loc_found := true;
		found_with_prob := found_with_prob + new_prob;
		attacker.curprob := new_prob;
	endif
	
case I_WAIT:
	if c = I_RECV then
		-- victim sends message
		change_state := send_msg(From_Resp,attacker.other,i, 1, victim.location); 
		put change_state;
		if change_state then 
			--find this conversation and update state
			for p: AttackerId do
				if (viclist[p].other = i) then viclist[p].state := R_SEND;
				endif;
			endfor;
		endif;
		-- attacker receives message
		change_state := recv_msg(From_Resp, attacker.other, i);
		put change_state;
		if change_state then 
			attacker.state := I_RECV;
			attacker.curprob := new_prob;
			attacker.locs_to_explore := attacker.locs_to_explore - 1;
		endif

		
	elsif c = I_KNOW then
		attacker.state := I_KNOW;
		loc_found := true;
		attacker.curprob := new_prob;
		found_with_prob := found_with_prob + new_prob;
	endif
case I_RECV:
	if c = I_SLEEP then
		attacker.state := I_SLEEP;
		attacker.curprob := new_prob;
		for p: AttackerId do
			if (viclist[p].other = i) then viclist[p].state := R_SLEEP;
			endif;
		endfor;
	elsif c = I_KNOW then 
		attacker.state := I_KNOW;
		attacker.curprob := new_prob;
		loc_found := true;
		found_with_prob := found_with_prob + new_prob;
	endif

case R_SLEEP:
	if c = R_SEND then
		--attacker sends Message
		change_state := send_msg(From_Resp, i, attacker.other, 1, attacker.location);
		if change_state then 
			attacker.state := R_SEND;
			attacker.curprob := new_prob;
		endif;
		--victim receives message
		change_state := recv_msg(From_Resp, i, attacker.other);
		if change_state then
			for p: AttackerId do
				if (viclist[p].other = i) then viclist[p].state := I_RECV;
				endif
			endfor;
		endif;

	elsif c = R_KNOW then 
		attacker.state := R_KNOW;
		attacker.curprob := new_prob;
		loc_found := true;
		found_with_prob := found_with_prob + new_prob;
	endif;

case R_SEND:
	if c = R_SLEEP then
		attacker.state := R_SLEEP;
		for p: AttackerId do
			if (viclist[p].other = i) then viclist[p].state := I_SLEEP;
			endif
		endfor
	elsif c = R_KNOW then 
		attacker.state := R_KNOW;
		attacker.curprob := new_prob;
		loc_found := true;
		found_with_prob := found_with_prob + new_prob;
	endif;
 endswitch;
endalias;
end;


--------------------------------------------------------------------------------
-- rules
--------------------------------------------------------------------------------
ruleset attacker: AttackerId do
	ruleset next_state : AgentStates do
		rule "next"
		(calc_prob(attacker, next_state) > 0) ==>
		begin
			step(attacker, next_state);
		end;
	end;
end;

--------------------------------------------------------------------------------
-- startstate
--------------------------------------------------------------------------------

startstate

  -- initialize attackers except other field
  undefine atkr;
  for i: AttackerId do
    atkr[i].role     := initiator;
	atkr[i].locs_to_explore := MaxLocations / NumAttackers;
	atkr[i].location := 1;			-- this location can be anything, attacker does not worry about revealing location
	atkr[i].state := I_SLEEP;
	atkr[i].other := 0;				-- every attacker is against 1 victim, which is agent id 0
	atkr[i].curprob := 1.0;
  end;

  -- initialize victim
  undefine vic;
		vic.location := 1;          -- since attacker starts to check from a very large number, they do not try 1 soon
		for k: AttackerId do
			vic.conversation_list[k].role := responder;
			vic.conversation_list[k].other := k;
			vic.conversation_list[k].state := R_SLEEP;
		endfor; 
  -- first attacker is in responder state, configure that
	atkr[1].role := responder;
	atkr[1].state := R_SLEEP;
	vic.conversation_list[1].role := initiator;
	vic.conversation_list[1].state := I_SLEEP;

-- initialize loc_found and found_with_prob
	undefine loc_found;
	undefine found_with_prob;
	loc_found := false;
	found_with_prob := 0.0;
		 
  -- initialize network 
  undefine net;
endstartstate;



--------------------------------------------------------------------------------
-- invariants
--------------------------------------------------------------------------------
--invariant "less than networksize outstanding messages in network"
--multisetcount (l:net, true) < NetworkSize;

invariant "loc_found with probability less than 0.0001"
found_with_prob < 0.00001
